/*
 *
 * HL demo implementation
 *
 * */

#include <random>
#include <atomic>
#include <iostream>
#include <sstream>

#include <hl/include/sfinae.hpp>
#include <hl/include/hl_ipc/conversion.hpp>
#include <hl/include/mpl_foreach.hpp>
#include <hl/include/hl_ipc/hl_ipc_specific.hpp>

#ifdef __J_ACTIVE
    #include <hl/include/hl_ipc/hl_json_backend.hpp>
#endif

#include <hl/include/hl_ipc/Interface.hpp>
#include <hl/include/linux/environment.hpp>
#include <hl/include/atomic_st_value.hpp>
#include <hl/include/register_type.hpp>

namespace impl
{
template <int > struct Services;

template <typename Class>
struct invoke_backend
{
    // template
    template <typename Backend>
    inline void
    invoke(Backend& backend)
    {
        std::cout << ((std::string)backend).c_str() << std::endl;
    }
};

template <typename StringType
        , typename Invoker>
struct archive_backend : Invoker
{
    // TODO we should implement any logic of you back end serialization
};


// Service base functionality
// implementation placed below
struct interface_first
{
    HL_META_CLASS_DECLARATION(interface_first);

    HL_RPC_TYPE(((std::string, first_filed), (std::string, second_field)), ReflType);
    HL_RPC_FUNCTION(
                ((bool, result))
                , run
                , ((std::string, first_field), (ReflType, refl_demo))
                );

    HL_RPC_FUNCTION(((bool, result)), exit, ((bool, mustexit)));

};

struct interface_second
{
    HL_META_CLASS_DECLARATION(interface_second);

    HL_RPC_FUNCTION(
                ((bool, result))
                , foo
                , ((std::string, first_field), (std::string, second_field))
                );

    HL_RPC_FUNCTION(((bool, result)), exit, ((bool, mustexit)));

};

#ifdef __J_ACTIVE
    struct implementation
            : hl::Interface<
                          hl::Impl<implementation, interface_first>
                        , std::string
                        , hl::JsonBackend<std::string, invoke_backend<implementation> > >
    {
        HL_RPC_SERVICE_CLIENT_DECLARATION(implementation, "I_am", "I_am", "");
    };
#else
    //  Certain implementation of service
    // with specified back end, we can use any Backend
    // , from textual information from binary, not important
    // generally we should implement Backend similar hl::JsonBackend e.g. and our new RPC will works with new format
    struct implementation_first
            : hl::Interface<
                          hl::Impl<implementation_first, interface_first>
                        , std::string
                        , impl::archive_backend<std::string, invoke_backend<implementation_first> > >
    {
        HL_RPC_SERVICE_CLIENT_DECLARATION(implementation_first, "I_am_first", "I_am_first", "");
    };

    //  Registering service to a compile time channel Services
    // we can extract this types by hl::manual_extract
    HL_REGISTER_TYPE(Services, implementation_first);

    struct implementation_second
            : hl::Interface<
                          hl::Impl<implementation_second, interface_second>
                        , std::string
                        , impl::archive_backend<std::string, invoke_backend<implementation_second> > >
    {
        HL_RPC_SERVICE_CLIENT_DECLARATION(implementation_second, "I_am_second", "I_am_second", "");
    };

    HL_REGISTER_TYPE(Services, implementation_second);
#endif

}

template <typename Type>
inline void
print_type()
{
    std::cout << __PRETTY_FUNCTION__ << std::endl;
}

//  The global test runner
// we should use this only as global
int main()
{
    typedef hl::manual_extract<impl::Services>::type services;
    print_type<services>();
    typedef hl::refl_extract<impl::interface_first::ftbl >::type functions;


    // but I can do anything here with methods
    print_type<functions>();
    std::cin.get();
    return 0;
}
